import os
import copy
import string
#from word2number import w2n

import stanza
from nltk.parse import CoreNLPParser
from nltk.parse.corenlp import CoreNLPDependencyParser

parser = CoreNLPParser(url='http://localhost:9000')
dep_parser = CoreNLPDependencyParser(url='http://localhost:9000')

nlp = stanza.Pipeline(
    "en",
    processors={"tokenize": "gum", "pos": "gum",
                "lemma": "gum", "depparse": "gum"},
    use_gpu=True,
    pos_batch_size=2000
)

# cd ./Desktop/Udep2Mono/NaturalLanguagePipeline/lib/stanford-corenlp-4.1.0
# java -mx4g -cp "*" edu.stanford.nlp.pipeline.StanfordCoreNLPServer -port 9000 -timeout 15000

replacement = {
    "out of": "out-of",
    "none of the": "none-of-the",
    "all of the": "all-of-the",
    "some of the": "some-of-the",
    "most of the": "most-of-the",
    "many of the": "many-of-the",
    "several of the": "several-of-the",
    "some but not all": "some-but-not-all"
}


def preprocess(sentence):
    replaced = {}
    processed = sentence.lower()
    for orig in replacement:
        if orig in processed:
            processed = processed.replace(orig, replacement[orig])
            replaced[replacement[orig]] = orig
    return processed, replaced


def dependencyParse(sentence, parser="stanford"):
    processed, replaced = preprocess(sentence)
    if parser == "stanford":
        return stanfordParse(processed), replaced
    elif parser == "stanza":
        return stanzaParse(processed), replaced


def stanzaParse(sentence):
    postag = {}
    words = {}
    parsed = nlp(sentence)
    parse_tree = []
    for sent in parsed.sentences:
        for word in sent.words:
            tree_node = postProcess(sent, word, postag, words)
            if len(tree_node) > 0:
                parse_tree.append(tree_node)

    # for tree_node in parse_tree:
    #    printTree(tree_node, postag, words)
    return parse_tree, postag, words


def postProcess(sent, word, postag, words):
    wordID = int(word.id)
    if wordID not in words:
        postag[word.text] = (wordID, word.xpos)
        words[wordID] = (word.text, word.xpos)
    if word.deprel != "punct":
        tree_node = [word.deprel, wordID,
                     word.head if word.head > 0 else "root"]
        return tree_node
    return []


def printTree(tree, tag, word):
    if tree[0] != "root":
        print(
            f"word: {word[tree[1]][0]}\thead: {word[tree[2]][0]}\tdeprel: {tree[0]}",
            sep="\n",
        )


def stanfordParse(sentence):
    postag = {}
    wordids = {}
    tokens = {}
    parse_tree = []

    tokenized = list(parser.tokenize(sentence))
    for i in range(len(tokenized)):
        tokens[tokenized[i]] = i+1

    parsed = list(dep_parser.raw_parse(sentence))[0]
    dep_rels = parsed.to_conll(4).split('\n')
    for dep_rel in dep_rels:
        rel = dep_rel.split('\t')
        if len(rel) == 4:
            dependent = tokens[rel[0]]
            govenor = int(rel[2])
            relation = rel[3].lower()
            postag[rel[0]] = (dependent, rel[1])
            wordids[dependent] = (rel[0], rel[1])
            if relation != "punct":
                parse_tree.append([relation, dependent, govenor])
    return parse_tree, postag, wordids



import os
from util import *
import numpy as np


dataset_path = "gold"
dataset = "gold"
path = os.path.join("../data", dataset_path)
in_name = "{}.polarized.txt".format(dataset)
data_name = "{}.txt".format(dataset)
data_path = os.path.join(path, data_name)
val_name = "{}.label.txt".format(dataset)
in_path = os.path.join(path, in_name)
val_path = os.path.join(path, val_name)
pos_name = "{}.pos.txt".format(dataset)
pos_path = os.path.join(path, pos_name)

num_matched = 0
num_total = 0

val = {}

with open(in_path, "r") as data:
    with open(val_path, "r") as annotation:
        with open(data_path, "r") as text:
            with open(pos_path, "r") as pos:
                lines = data.readlines()
                annotations_val = annotation.readlines()
                sentences = text.readlines()
                postags = pos.readlines()

                for i in range(len(sentences)):
                    annotation_val = annotations_val[i]
                    sentence = sentences[i]
                    val[sentence] = annotation_val

                for i in range(0, len(lines), 2):
                    sent = lines[i]
                    postag = postags[i+1].split(" ")
                    vec = convert2vector(lines[i+1])
                    vec_val = convert2vector(val[sent])

                    if len(vec) == len(vec_val):
                        for j in range(len(vec)):
                            if postag[j] in ["NN", "NNS", "DT", "VB",
                                             "VBD", "VBN", "VBG", "VBP",
                                             "VBZ", "DT", "CD", "JJ", "RB",
                                             "RBR", 'RBS']:
                                num_total += 1
                                if vec[j] == vec_val[j]:
                                    num_matched += 1

print("Number of tokens: ", num_total)
print("Number of correct tokens: ", num_matched)
print("Percentage: ", (num_matched)/num_total)



class BinaryDependencyTree:
    def __init__(self, val, left, right, key, wid=None, npos=None):
        self.val = val
        self.parent = ""
        self.left = left
        self.right = right
        self.mark = "0"
        self.id = wid
        self.npos = npos
        self.key = key

    def isTree(self):
        return not (type(self.left) is str and type(self.right) is str)

    def getVal(self):
        return self.val

    def getLeft(self):
        return self.left

    def getRight(self):
        return self.right


class Binarizer:
    def __init__(self, parseTable=None, postag=None, words=None):
        self.postag = postag
        self.parseTable = parseTable
        self.words = words
        self.id = 0

    def process_not(self, children):
        if len(children) > 1:
            if children[0][0] == "advmod":
                if self.words[children[1][1]][0] == "not":
                    return [children[1]]
        return children

    def compose(self, head):
        children = list(filter(lambda x: x[2] == head, self.parseTable))
        children.sort(
            key=(lambda x: (x[2]-x[1])*20 if x[2] > x[1] else abs(x[2]-x[1])), reverse=True)
        children = self.process_not(children)
        if len(children) == 0:
            word = self.words[head][0]
            tag = self.words[head][1]
            binaryTree = BinaryDependencyTree(
                word, "N", "N", self.id, head, tag)
            self.id += 1
            return binaryTree, [binaryTree.key]
        else:
            topDep = children[0]
        self.parseTable.remove(topDep)

        left, left_rel = self.compose(topDep[1])
        right, right_rel = self.compose(topDep[2])
        binaryTree = BinaryDependencyTree(topDep[0], left, right, self.id)

        binaryTree.left.parent = binaryTree
        binaryTree.right.parent = binaryTree

        left_rel.append(binaryTree.key)
        self.id += 1
        return binaryTree, left_rel + right_rel

    def binarization(self):
        self.id = 0
        self.relation = []
        root = list(filter(lambda x: x[0] == "root", self.parseTable))[0][1]
        return self.compose(root)



import heapq
from pattern.en import conjugate

relations = [
    "acl",
    "acl:relcl",
    "advcl",
    "advmod",
    "advmod:count",
    "amod",
    "appos",
    "aux",
    "aux:pass",
    "case",
    "cc",
    "cc:preconj",
    "ccomp",
    "clf",
    "compound",
    "compound:prt",
    "conj",
    "cop",
    "csubj",
    "csubj:pass",
    "dep",
    "det",
    "det:predet",
    "discourse",
    "dislocated",
    "expl",
    "fixed",
    "flat",
    "goeswith",
    "iobj",
    "list",
    "mark",
    "nmod",
    "nmod:poss",
    "nmod:npmod",
    "nmod:tmod",
    "nmod:count",
    "nsubj",
    "nsubj:pass",
    "nummod",
    "obj",
    "obl",
    "obl:npmod",
    "obl:tmod",
    "orphan",
    "parataxis",
    "punct",
    "reparandum",
    "root",
    "vocative",
    "xcomp"
]

negate_mark = {
    "+": "-",
    "-": "+",
    "=": "="
}

det_mark = {
    "det:univ": ("+", "-"),
    "det:exist": ("+", "+"),
    "det:limit": ("+", "="),
    "det:negation": ("+", "-")
}

det_type_words = {
    "det:univ": ["all", "every", "each", "any"],
    "det:exist": ["a", "an", "some", "double", "triple"],
    "det:limit": ["such", "both", "the", "this", "that",
                  "those", "these", "my", "his", "her",
                  "its", "either", "both", "another"],
    "det:negation": ["no", "neither", "never", "few"]
}

negtive_implicative = ["refuse", "reject", "oppose", "forget",
                       "hesitate", "without", "disapprove", "disagree",
                       "eradicate", "erase", "dicline", "eliminate",
                       "decline", "resist", "block", "stop", "hault",
                       "disable", "disinfect", "disapear", "disgard",
                       "disarm", "disarrange", "disallow", "discharge",
                       "disbelieve", "disclaim", "disclose", "disconnect",
                       "disconnect", "discourage", "discredit", "discorporate",
                       "disengage", "disentangle", "dismiss", "disobeye",
                       "distrust", "disrupt", "suspen", "suspend ",
                       "freeze", "remove", "regret", "object", "impossible",
                       "hate"
                       ]

at_least_implicative = ["smoke", "for", "buy", "drink", "take", "hold", "receive",
                        "get", "catch"]

exactly_implicative = ["like", "love", "admires", "marry"]


def is_implicative(verb, imp_type):
    return conjugate(verb, tense="present", person=1, number="singular") in imp_type


def det_type(word):
    for det in det_type_words:
        if word.lower() in det_type_words[det]:
            return det


arrows = {
    "+": "\u2191",
    "-": "\u2193",
    "=": "=",
    "0": ""
}


def btreeToList(binaryDepdency, length, replaced, verbose=2):
    annotated = []
    postags = []
    reverse = {}

    def toList(tree):
        treelist = []
        if tree.getVal() not in relations:
            treelist.append(tree.npos)
            if tree.getVal() == "n't":
                tree.val = "not"
            if tree.val in replaced.keys():
                original = replaced[tree.val].split()
                new = []
                for word in original:
                    new.append(word + arrows[tree.mark])
                reverse[tree.val + arrows[tree.mark]
                        ] = '%s' % ', '.join(map(str, new)).replace(",", " ")
            word = tree.getVal() + arrows[tree.mark]
            if verbose == 2:
                word += str(tree.key)
            index = tree.id
            heapq.heappush(annotated, (int(index), word))
            heapq.heappush(postags, (int(index), tree.npos))
            treelist.append(word)
        else:
            word = tree.getVal() + arrows[tree.mark]
            if verbose == 2:
                word += str(tree.key)
            treelist.append(word)

        left = tree.getLeft()
        right = tree.getRight()

        if left is not 'N':
            treelist.append(toList(left))

        if right is not 'N':
            treelist.append(toList(right))

        return treelist
    return toList(binaryDepdency), annotated, postags, reverse


def convert2vector(result):
    result_vec = []
    if type(result) is "str":
        result_ls = result.split()
    else:
        result_ls = result
    for word in result_ls:
        if arrows['+'] in word:
            result_vec.append(1)
        elif arrows['-'] in word:
            result_vec.append(-1)
        elif arrows['='] in word:
            result_vec.append(0)
    return result_vec



import os
import argparse

datasets = ["gold", "diagnostic",
            "monotonicity_hard", "monotonicity_simple",
            "boolean", "conditional", "counting", "negation"]


def parse_dataset_path(dataset):
    if dataset == "gold":
        dataset_path = "gold"
    elif dataset == "diagnostic":
        dataset_path = "GLUE/glue_data/diagnostic"
    elif dataset == "monotonicity_hard":
        dataset_path = "SEG/monotonicity_hard"
    elif dataset == "monotonicity_simple":
        dataset_path = "SEG/monotonicity_simple"
    elif dataset == "boolean":
        dataset_path = "SEG/boolean"
    elif dataset == "conditional":
        dataset_path = "SEG/conditional"
    elif dataset == "counting":
        dataset_path = "SEG/counting"
    elif dataset == "negation":
        dataset_path = "SEG/negation"

    in_name = "{}.txt".format(dataset)
    val_name = "{}.label.txt".format(dataset)
    out_name = "{}.polarized.txt".format(dataset)
    incorrect_log_name = "{}.incorrect.txt".format(dataset)
    unmatched_log_name = "{}.unmatched.txt".format(dataset)
    unmatched_val_name = "{}.unmatched.ccg.txt".format(dataset)
    except_log_name = "{}.exception.txt".format(dataset)
    pos_name = "{}.pos.txt".format(dataset)

    path = os.path.join("../data", dataset_path)
    in_path = os.path.join(path, in_name)
    val_path = os.path.join(path, val_name)
    out_path = os.path.join(path, out_name)
    incorrect_log_path = os.path.join(path, incorrect_log_name)
    unmatched_log_path = os.path.join(path, unmatched_log_name)
    unmatched_val_path = os.path.join(path, unmatched_val_name)
    exception_log_path = os.path.join(path, except_log_name)
    pos_path = os.path.join(path, pos_name)

    return {
        "in_path": in_path,
        "val_path": val_path,
        "out_path": out_path,
        "incorrect_log_path": incorrect_log_path,
        "unmatched_log_path": unmatched_log_path,
        "unmatched_val_path": unmatched_val_path,
        "exception_log_path": exception_log_path,
        "pos_path": pos_path
    }


def polarize_dataset(dataset, add_pos=0, mode="polarize", parser="stanza"):
    path = parse_dataset_path(dataset)

    if mode == "polarize":
        with open(path['in_path'], "r") as data:
            sentences = data.readlines()
            annotations, exceptioned = run_polarize_pipeline(
                sentences,
                verbose=2,
                parser=parser)

            with open(path['out_path'], "w") as correct:
                for sent in annotations:
                    correct.write("%s\n" % sent[0])

    elif mode == "eval":
        with open(path['in_path'], "r") as data:
            with open(path['val_path'], "r") as annotation:
                lines = data.readlines()
                annotations_val = annotation.readlines()
                annotations, exceptioned, incorrect = polarize_eval(
                    lines, annotations_val,
                    verbose=2,
                    parser="stanford")

                with open(path['out_path'], "w") as correct:
                    for sent in annotations:
                        correct.write("%s\n" % sent["annotated"])

                with open(path['incorrect_log_path'], "w") as incorrect_log:
                    with open(path['unmatched_val_path'], "w") as unmatched_val:
                        for sent in incorrect:
                            incorrect_log.write(sent[0])
                            unmatched_val.write(sent[2])

                with open(path['exception_log_path'], "w") as except_log:
                    for sent in exceptioned:
                        except_log.write(sent[0])
                        except_log.write(sent[1])

                if add_pos == 1:
                    with open(path['pos_path'], "w") as postag:
                        for sent in annotations:
                            postag.write(sent["orig"])
                            postag.write("%s\n" % sent["postag"])

        with open(path['incorrect_log_path'], "r") as incorrect_log:
            with open(path['unmatched_val_path'], "r") as annotation:
                lines = incorrect_log.readlines()
                annotations_val = annotation.readlines()
                annotations, exceptioned, incorrect = polarize_eval(
                    lines, annotations_val, verbose=2, parser="stanza"
                )

                with open(path['out_path'], "a") as correct:
                    for sent in annotations:
                        correct.write("%s\n" % sent["annotated"])

                with open(path['unmatched_log_path'], "w") as unmatched_log:
                    for sent in incorrect:
                        unmatched_log.write(sent[0])
                        unmatched_log.write("%s\n" % sent[1])
                        unmatched_log.write(sent[2])
                        if add_pos == 1:
                            unmatched_log.write("%s\n" % sent[3])
                        unmatched_log.write("\n")

                with open(path['exception_log_path'], "a") as except_log:
                    for sent in exceptioned:
                        except_log.write(sent[0])
                        except_log.write(sent[1])

                if add_pos == 1:
                    with open(path['pos_path'], "a") as postag:
                        for sent in annotations:
                            postag.write(sent["orig"])
                            postag.write("%s\n" % sent["postag"])


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--dataset", help="The name of a working or evaluation dataset", type=str, default="gold")
    parser.add_argument(
        "--pos", help="Print part-of-speech tags also", type=int, default=0)
    parser.add_argument(
        "--mode", help="The pipeline mode: [1] polarize [2] eval", type=str, default="polarize")
    parser.add_argument(
        "--parser", help="The dependency parser used: [1] stanford [2] stanza", type=str, default="stanza")
    args = parser.parse_args()

    from udep2mono.polarization import polarize_eval, run_polarize_pipeline
    polarize_dataset(args.dataset, args.pos, args.mode, args.parser)



relationPriority = {
    "fixed": 9,
    "goeswith": 9,
    "parataxis": 9,
    "appos": 9,
    "amod": 10,
    "compound": 10,
    "compound:prt": 10,
    "nummod": 10,
    "expl": 10,
    "det": 15,
    "det:predet": 15,
    "cc:preconj": 25,
    "cc": 25,
    "acl:relcl": 16,
    "acl": 16,
    "cop": 40,
    "obj": 40,
    "case": 20,
    "mark": 15,
    "xcomp": 10,
    "nmod:poss": 10,
    "nmod:npmod": 40,
    "nmod:tmod": 40,
    "nmod:count": 40,
    "nmod": 40,
    "obl": 40,
    "iobj": 40,
    "obl:npmod": 40,
    "obl:tmod": 42,
    "advmod": 20,
    "advmod:count": 20,
    "advcl": 40,
    "ccomp": 10,
    "aux:pass": 45,
    "aux": 45,
    "dep": 45,
    "conj": 45,
    "conj:vp": 45,
    "csubj": 60,
    "csubj:pass": 60,
    "nsubj": 60,
    "nsubj:pass": 60,
    "nsubj:wh": 60,
    "conj:sent": 80,
    "punct": 90,
    "root": 90,
    "discourse": 80
}



import numpy as np
from tqdm import tqdm
from copy import deepcopy
from udep2mono.binarization import Binarizer, BinaryDependencyTree
from udep2mono.dependency_parse import dependencyParse
from udep2mono.util import *


class Polarizer:
    def __init__(self, dependtree=None, relation=None):
        self.dependtree = dependtree
        self.sentence_head = []
        self.relation = relation
        self.polarize_function = {
            "acl": self.polarize_acl_relcl,
            "acl:relcl": self.polarize_acl_relcl,
            "advcl": self.polarize_acl_relcl,
            "advmod": self.polarize_advmod,
            "advmod:count": self.polarize_advmod,
            "amod": self.polarize_amod,
            "appos": self.polarize_inherite,
            "aux": self.polarize_aux,
            "aux:pass": self.polarize_aux,
            "case": self.polarize_case,
            "cc": self.polarize_cc,
            "cc:preconj": self.polarize_det,
            "ccomp": self.polarize_ccomp,
            "compound": self.polarize_inherite,
            "compound:prt": self.polarize_inherite,
            "conj": self.polarize_inherite,
            "cop": self.polarize_inherite,
            "csubj": self.polarize_nsubj,
            "csubj:pass": self.polarize_nsubj,
            "dep": self.polarize_dep,
            "det": self.polarize_det,
            "det:predet": self.polarize_det,
            "discourse": self.polarize_discourse,
            "expl": self.polarize_expl,
            "fixed": self.polarize_inherite,
            "flat": self.polarize_inherite,
            "goeswith": self.polarize_inherite,
            "iobj": self.polarize_inherite,
            "mark": self.polarize_inherite,
            "nmod": self.polarize_nmod,
            "nmod:npmod": self.polarize_nmod,
            "nmod:tmod": self.polarize_nmod,
            "nmod:poss": self.polarize_nmod_poss,
            "nsubj": self.polarize_nsubj,
            "nsubj:pass": self.polarize_nsubj,
            "nummod": self.polarize_nummod,
            "obj": self.polarize_obj,
            "obl": self.polarize_obj,
            "obl:npmod": self.polarize_oblnpmod,
            "obl:tmod": self.polarize_inherite,
            "parataxis": self.polarize_inherite,
            "xcomp": self.polarize_obj,
        }
        self.treeLog = []
        self.polarLog = []

    def polarize_deptree(self):
        self.polarize(self.dependtree)

    def polarize(self, tree):
        if tree.isTree():
            self.polarize_function[tree.val](tree)

    def polarize_acl_relcl(self, tree):
        self.polarLog.append("polarize_acl:relcl")
        self.sentence_head.append(tree)
        right = tree.getRight()
        left = tree.getLeft()

        if tree.mark != "0":
            right.mark = tree.mark
        else:
            tree.mark = "+"
            right.mark = "+"

        left.mark = "+"
        if left.isTree():
            self.polarize(left)
        if right.isTree():
            self.polarize(right)

        tree.mark = right.mark

        if right.mark == "-" and left.npos != "VBD":
            self.negate(left, -1)
        elif right.mark == "=" and left.npos != "VBD":
            self.equalize(left)
        elif right.val == "impossible":
            self.negate(left, -1)

        self.sentence_head.pop()

    def polarize_advmod(self, tree):
        self.polarLog.append("advmod")

        left = tree.getLeft()
        right = tree.getRight()
        self.polarize_inherite(tree)
        root_mark = tree.mark

        more_than = False
        less_than = False
        at_most = False
        at_least = False

        if tree.left.val == "fixed":
            more_than = tree.left.right.val.lower(
            ) == "more" and tree.left.left.val.lower() == "than"
            less_than = tree.left.right.val.lower(
            ) == "less" and tree.left.left.val.lower() == "than"
            at_most = tree.left.left.val.lower() == "most" and tree.left.right.val.lower() == "at"
            at_least = tree.left.left.val.lower() == "least" and tree.left.right.val.lower() == "at"

        if tree.left.val == "case":
            at_most = tree.left.left.val.lower() == "at" and tree.left.right.val.lower() == "most"
            at_least = tree.left.left.val.lower() == "at" and tree.left.right.val.lower() == "least"

        if left.val.lower() in ["many", "most"]:
            right.mark = "="
            if isinstance(tree.parent, BinaryDependencyTree) and tree.parent.val == "amod":
                self.equalize(tree.parent.right)
        elif left.val.lower() in ["not", "no", "n’t", "never"]:
            self.negate(right, -1)
        elif left.val.lower() in ["exactly"]:
            self.equalize(tree.parent.parent)
            left.mark = root_mark
            tree.mark = right.mark
        elif more_than:
            if right.val == "nummod":
                right.left.mark = "-"
            elif right.npos == "CD":
                right.mark = "-"
            elif right.isTree() and right.left.val == "nummod":
                right.left.left.mark = "-"
                right.left.right.mark = "+"
        elif less_than:
            self.negate(right, self.relation.index(tree.key))
            if right.val == "nummod":
                right.left.mark = "+"
            elif right.npos == "CD":
                right.mark = "+"
            elif right.isTree() and right.left.val == "nummod":
                right.left.left.mark = "+"
                right.left.right.mark = "-"
        elif at_most:
            self.negate(right, self.relation.index(tree.key))
            if right.val == "nmod":
                right.right.mark = "+"
            elif right.val == "nummod":
                right.left.mark = "+"
            elif right.npos == "CD":
                right.mark = "+"
            elif right.isTree() and right.left.val == "nummod":
                right.left.left.mark = "+"
                right.left.right.mark = "-"
        elif at_least:
            if right.val == "nmod":
                right.right.mark = "-"
            elif right.val == "nummod":
                right.left.mark = "-"
            elif right.npos == "CD":
                right.mark = "-"
            elif right.isTree() and right.left.val == "nummod":
                right.left.left.mark = "-"
                right.left.right.mark = "+"

        if left.val.lower() == "when":
            self.equalize(self.dependtree)

    def polarize_amod(self, tree):
        self.polarLog.append("amod")

        left = tree.getLeft()
        right = tree.getRight()
        self.polarize_inherite(tree)

        at_most = False
        at_least = False

        if isinstance(tree.parent, BinaryDependencyTree):
            at_most = tree.parent.left.val.lower() == "at" and left.val.lower() == "most"
            at_least = tree.parent.left.val.lower() == "at" and left.val.lower() == "least"

        if at_most:
            self.negate(right, self.relation.index(tree.key))
            if right.val == "nummod":
                right.left.mark = "+"
            elif right.npos == "CD":
                right.mark = "+"
            elif right.isTree() and right.left.val == "nummod":
                right.left.left.mark = "+"
                right.left.right.mark = "-"
        elif at_least:
            if right.val == "nummod":
                right.left.mark = "-"
            elif right.npos == "CD":
                right.mark = "-"
            elif right.isTree() and right.left.val == "nummod":
                right.left.left.mark = "-"
                right.left.right.mark = "+"
        elif left.val.lower() in ["many", "most"] and not at_most:
            self.equalize(right)
            tree.mark = right.mark
        elif left.val.lower() == "fewer":
            self.nounMarkReplace(right, "-")
        elif left.val == "advmod":
            if left.right.val == "many":
                self.equalize(right)
                tree.mark = right.mark
            if left.left.val.lower() == "not":
                self.top_down_negate(
                    tree, "amod", self.relation.index(tree.key))

    def polarize_aux(self, tree):
        self.polarize_inherite(tree)

    def polarize_case(self, tree):
        self.polarLog.append("polarize_case")

        right = tree.getRight()
        left = tree.getLeft()
        self.polarize_inherite(tree)

        if tree.mark != "0":
            left.mark = tree.mark
            right.mark = tree.mark
        else:
            tree.mark = "+"
            left.mark = "+"
            right.mark = "+"

        if left.val == "without":
            if right.isTree():
                self.polarize(right)
            self.negate(tree, self.relation.index(left.key))
        elif right.npos == "CD":
            right.mark = "="
            if left.isTree():
                self.polarize(left)
        # elif left.val.lower() == "for":
        #    if right.isTree():
        #        self.polarize(right)
        #    self.equalize(right)
        elif right.val == "nmod:poss":
            left.mark = "="
            if right.isTree():
                self.polarize(right)
        elif left.val == "except":
            right.mark = "="
            self.polarize(right)

    def polarize_cc(self, tree):
        right = tree.getRight()
        left = tree.getLeft()

        left.mark = "+"
        right.mark = "+"

        if right.val != "expl" and right.val != "det":
            right.mark = tree.mark

        if right.isTree():
            self.polarize(right)

        if left.id == 1:
            self.equalize(right)

    def polarize_ccomp(self, tree):
        right = tree.getRight()
        left = tree.getLeft()

        if tree.mark != "0":
            right.mark = tree.mark

        if right.isTree():
            self.polarize(right)

        left.mark = right.mark
        if left.isTree():
            self.polarize(left)

    def polarize_dep(self, tree):
        self.polarLog.append("polarize_dep")

        right = tree.getRight()
        left = tree.getLeft()

        if tree.mark != "0":
            right.mark = tree.mark
            left.mark = tree.mark
        else:
            tree.mark = "+"
            right.mark = "+"
            left.mark = "+"

        if right.isTree():
            self.polarize(right)

        if left.isTree():
            self.polarize(left)

    def polarize_det(self, tree):
        self.polarLog.append("polarize_det")
        right = tree.getRight()
        left = tree.getLeft()

        if tree.mark != "0":
            left.mark = tree.mark
            right.mark = tree.mark
        else:
            left.mark = "+"
            right.mark = "+"

        detType = det_type(left.val)
        if detType is None:
            detType = "det:exist"

        if left.val.lower() == "any":
            if isinstance(tree.parent, BinaryDependencyTree):
                if isinstance(tree.parent.parent, BinaryDependencyTree):
                    negate_signal = tree.parent.parent.left
                    if negate_signal.val == "not":
                        detType = "det:exist"
                    if negate_signal.val == "det" and negate_signal.left.val.lower() == "no":
                        detType = "det:exist"
        detMark = det_mark[detType]

        right.mark = detMark[1]
        tree.mark = detMark[1]

        if right.isTree():
            self.polarize(right)

        if detType == "det:negation":
            self.top_down_negate(tree, "det", self.relation.index(tree.key))

    def polarize_discourse(self, tree):
        self.polarize_inherite(tree)
        right = tree.getRight()
        left = tree.getLeft()

        more_than = False
        less_than = False

        if tree.left.val == "fixed":
            more_than = left.right.val.lower() == "more" and left.left.val.lower() == "than"
            less_than = left.right.val.lower() == "less" and left.left.val.lower() == "than"

        if more_than:
            if right.val == "nummod":
                right.left.mark = "-"
            elif right.npos == "CD":
                right.mark = "-"
            elif right.isTree() and right.left.val == "nummod":
                right.left.left.mark = "-"
                right.left.right.mark = "+"
        elif less_than:
            self.negate(right, self.relation.index(tree.key))
            if right.val == "nummod":
                right.left.mark = "+"
            elif right.npos == "CD":
                right.mark = "+"
            elif right.isTree() and right.left.val == "nummod":
                right.left.left.mark = "+"
                right.left.right.mark = "-"

    def polarize_expl(self, tree):
        right = tree.getRight()
        left = tree.getLeft()

        left.mark = "+"
        right.mark = "+"

        if self.dependtree.left.mark == "-":
            right.mark = "-"

        if left.isTree():
            self.polarize(left)

        if right.isTree():
            self.polarize(right)

    def polarize_nmod(self, tree):
        right = tree.getRight()
        left = tree.getLeft()

        if tree.mark != "0":
            right.mark = tree.mark
        else:
            right.mark = "+"
        left.mark = "+"

        if right.npos == "DT" or right.npos == "CC":
            detType = det_type(right.val)
            if detType == None:
                detType = "det:exist"
            left.mark = det_mark[detType][1]
            if detType == "det:negation":
                self.top_down_negate(
                    tree, "nmod", self.relation.index(tree.key))
        elif right.val.lower() in ["many", "most"]:
            left.mark = "="

        if left.isTree():
            self.polarize(left)

        if right.isTree():
            self.polarize(right)

        if left.val == "case":
            if isinstance(tree.parent, BinaryDependencyTree):
                if tree.parent.left.val.lower() == "more":
                    left.right.mark = "-"

        tree.mark = right.mark
        if right.mark == "-":
            self.negate(left, -1)
        elif right.mark == "=":
            self.equalize(left)

    def polarize_nmod_poss(self, tree):
        right = tree.getRight()
        left = tree.getLeft()

        left.mark = tree.mark
        if left.isTree():
            self.polarize(left)
        else:
            left.mark = "+"

        right.mark = tree.mark
        if self.searchDependency("det", tree.left):
            right.mark = left.mark
        if right.isTree():
            self.polarize(right)
        else:
            right.mark = "+"

    def polarize_nsubj(self, tree):
        self.polarLog.append("polarize_nsubj")
        right = tree.getRight()
        left = tree.getLeft()

        if tree.mark != "0":
            right.mark = tree.mark
            left.mark = tree.mark
        else:
            tree.mark = "+"
            right.mark = "+"
            left.mark = "+"

        if self.searchDependency("expl", right):
            self.polarize(left)
            self.polarize(right)
            return

        self.polarize(right)

        if left.val.lower() == "that":
            self.equalize(right)
        if not isinstance(tree.parent, str):
            if tree.parent.left.val.lower() == "that":
                self.equalize(left)

        if left.isTree():
            self.polarize(left)
        else:
            if left.val.lower() in ["nobody"]:
                self.negate(tree, self.relation.index(tree.key))

        if tree.mark == "0":
            tree.mark = right.mark

        if left.npos == "NN":
            left.mark = tree.mark

        if is_implicative(right.val.lower(), negtive_implicative):
            tree.mark = "-"

    def polarize_nummod(self, tree):
        self.polarLog.append("polarize_nummod")
        right = tree.getRight()
        left = tree.getLeft()

        left.mark = "="
        if tree.mark != "0":
            right.mark = tree.mark
        else:
            right.mark = "+"

        if tree.parent == "compound":
            right.mark = left.mark

        if left.isTree():
            if left.val == "advmod":
                left.mark = "+"
            self.polarize(left)
            if left.mark == "=":
                right.mark = left.mark
                tree.mark = left.mark
        elif left.id == 1:
            left.mark = "="
            # right.mark = "="

        if not isinstance(tree.parent, str):
            if is_implicative(tree.parent.right.val, at_least_implicative):
                left.mark = "-"
                # right.mark = "-"
            elif is_implicative(tree.parent.right.val, exactly_implicative):
                left.mark = "="
                # right.mark = "="

        if right.isTree():
            self.polarize(right)

    def polarize_obj(self, tree):
        right = tree.getRight()
        left = tree.getLeft()

        if tree.mark != "0":
            right.mark = tree.mark
        else:
            right.mark = "+"
            tree.mark = "+"
        left.mark = "+"

        if right.isTree():
            self.polarize(right)

        if left.isTree():
            self.polarize(left)

        if is_implicative(right.val.lower(), negtive_implicative):
            tree.mark = "-"
            self.negate(left, -1)

        if left.val == "mark" and left.left.val == "to":
            left.left.mark = right.mark

    def polarize_obl(self, tree):
        self.polarLog.append("polarize_obl")

        right = tree.getRight()
        left = tree.getLeft()

        if tree.mark != "0":
            right.mark = tree.mark
        else:
            right.mark = "+"
        left.mark = "+"

        if right.isTree():
            self.polarize(right)

        if left.isTree():
            self.polarize(left)

        if right.mark == "-":
            self.negate(left, -1)

    def polarize_oblnpmod(self, tree):
        right = tree.getRight()
        left = tree.getLeft()

        if left.isTree():
            self.polarize(left)
        right.mark = left.mark
        if right.isTree():
            self.polarize(right)

    def polarize_inherite(self, tree):
        right = tree.getRight()
        left = tree.getLeft()

        if tree.mark != "0":
            right.mark = tree.mark
            left.mark = tree.mark
        else:
            right.mark = "+"
            left.mark = "+"

        if right.isTree():
            self.polarize(right)

        if left.val.lower() == "there":
            left.mark = "+"

        if left.isTree():
            self.polarize(left)
        elif left.val.lower() == "if":
            self.negate(right, -1)

    def searchDependency(self, deprel, tree):
        if tree.val == deprel:
            return True
        else:
            right = tree.getRight()
            left = tree.getLeft()

            leftFound = False
            rightFound = False

            if not isinstance(right, str) and right.isTree():
                rightFound = self.searchDependency(deprel, right)

            if not isinstance(left, str) and left.isTree():
                leftFound = self.searchDependency(deprel, left)

            return leftFound or rightFound

    def verbMarkReplace(self, tree, mark):
        if isinstance(tree, str):
            return
        if tree.npos is not None and "VB" in tree.npos:
            tree.mark = mark
        self.verbMarkReplace(tree.left, mark)
        self.verbMarkReplace(tree.right, mark)

    def nounMarkReplace(self, tree, mark):
        if isinstance(tree, str):
            return False
        if tree.npos is not None and "NN" in tree.npos:
            tree.mark = mark
            return True
        right = self.nounMarkReplace(tree.right, mark)
        if not right:
            self.nounMarkReplace(tree.left, mark)

    def equalize(self, tree):
        if tree.isTree():
            self.equalize(tree.getRight())
            self.equalize(tree.getLeft())
            if tree.mark != "0":
                tree.mark = "="
        else:
            if tree.npos != "CC" and tree.val.lower() != "when":
                tree.mark = "="

    def negate_condition(self, tree, anchor):
        not_truth_connection = not tree.val in ["and", "or"]
        not_empty_mark = tree.mark != "0"
        return not_empty_mark and not_truth_connection

    def top_down_negate(self, tree, deprel, anchor):
        if not isinstance(tree.parent, BinaryDependencyTree):
            return
        if tree.parent.left.val == deprel:
            self.negate(tree.parent.left, anchor)
            self.negate(tree.parent.right, -1)
        elif tree.parent.right.val == deprel:
            self.negate(tree.parent.right, anchor)
            self.negate(tree.parent.left, -1)

    def negate(self, tree, anchor):
        if isinstance(tree, str):
            return
        if tree.val == "cc" and tree.right.val in ["expl", "nsubj", "det"]:
            return
        if tree.isTree():
            # print(tree.val)
            if self.relation.index(tree.key) > anchor or "nsubj" in tree.val:
                # print(tree.val)
                self.negate(tree.getRight(), anchor)
                self.negate(tree.getLeft(), anchor)
                if self.negate_condition(tree, anchor):
                    tree.mark = negate_mark[tree.mark]
        else:
            if self.relation.index(tree.key) > anchor and self.negate_condition(tree, anchor):
                if tree.npos != "EX":
                    # print(tree.val)
                    tree.mark = negate_mark[tree.mark]


def run_polarize_pipeline(sentences, verbose=0, parser="stanza"):
    binarizer = Binarizer()
    polarizer = Polarizer()

    annotations = []
    exceptioned = []

    for i in tqdm(range(len(sentences))):
        # Universal Dependency Parse
        sent = sentences[i]

        if len(sent) == 0:
            continue

        parsed, replaced = dependencyParse(sent, parser)
        tree, postag, words = parsed

        # print("s")
        # print(words)
        # print(parseTreeCopy)

        # Binarization
        binarizer.parseTable = tree
        binarizer.postag = postag
        binarizer.words = words
        sexpression = ""
        annotated = ""

        try:
            binaryDepdency, relation = binarizer.binarization()

            if verbose == 2:
                sexpression, annotated, _, _ = btreeToList(
                    binaryDepdency, len(words), replaced, 0)
                sexpression = '[%s]' % ', '.join(
                    map(str, sexpression)).replace(",", " ")
        except Exception as e:
            if verbose == 2:
                print(str(e))
            exceptioned.append(sent)
            continue

        # Polarization
        polarizer.dependtree = binaryDepdency
        polarizer.relation = relation
        try:
            polarizer.polarize_deptree()
            polarized, annotated, postags, reverse = btreeToList(
                binaryDepdency, len(words), replaced, 0)
            polarized = '[%s]' % ', '.join(
                map(str, polarized)).replace("'", "")
            polarized = polarized.replace(",", "")
        except Exception as e:
            if verbose == 2:
                print(str(e))
            print(sent)
            exceptioned.append(sent)
            continue

        # Postprocessing
        result = []
        while annotated:
            next_item = heapq.heappop(annotated)
            result.append(next_item[1])
        result = '%s' % ', '.join(map(str, result)).replace(",", "")
        for word in reverse:
            result = result.replace(word, reverse[word])
        annotations.append(
            (result, sent, polarized, postags, polarizer.dependtree))

    return annotations, exceptioned


def polarize_eval(sentences, annotations_val=[], verbose=0, parser="stanza"):
    num_unmatched = 0
    incorrect = []
    annotations = []

    outputs, exceptioned = run_polarize_pipeline(sentences, verbose, parser)

    for i in tqdm(range(len(outputs))):
        output = outputs[i]

        pos = []
        postags = output[3]
        while postags:
            next_item = heapq.heappop(postags)
            pos.append(next_item[1])

        pos = '%s' % ', '.join(map(str, pos)).replace(",", "")

        vec = convert2vector(output[0])
        if len(annotations_val) > 0:
            annotation_val = annotations_val[i]
            vec_val = convert2vector(annotation_val)
            if len(vec) == len(vec_val):
                if not np.array_equal(vec, vec_val):
                    num_unmatched += 1
                    incorrect.append(
                        (output[1], output[0], annotation_val, pos))
                if parser == "stanford":
                    continue

        validate = ""
        if len(annotations_val) > 0:
            validate = annotations_val[i]

        annotations.append(
            {
                "annotated": output[0],
                "polarized": output[2],
                "validation": validate,
                "orig": output[1],
                "postag": pos
            }
        )

    print()
    print("Number of unmatched sentences: ", num_unmatched)
    return annotations, exceptioned, incorrect


if __name__ == '__main__':
    sentences = ["At most 6 dogs are hungry"]
    annotations, _ = run_polarize_pipeline(
        sentences, verbose=2, parser="stanford")
    print(annotations[0][0])
    print(annotations[0][2])
    print(annotations[0][3])
